(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

module Make(V:Constant.S) : XXXCompile.S =
struct

  module PPC = PPCArch.Make(V)
  include CompileCommon.Make(PPC) 

  let do_sta = !Config.sta

(* PPO *)
  open C
  open E
  open R
    
  let dp f = f (Dp (Diff,Irr))      
  and dpr f = f (Dp (Diff,Dir R))      
  and dpw f = f (Dp (Diff,Dir W))      
  and ctrl f = f (Ctrl (Diff,Irr))
  and ctrlw f = f (Ctrl (Diff,Dir W))
  and ctrlr f = f (Ctrl (Diff,Dir R))
  and poswr f = f (Po(Same,Dir W,Dir R))

  let cons r rs = r (fun r -> r::rs)

  let single f r = f (ER r)
  let seq rs f = f (ERS rs)

  let do_ppo f k =
    let k = dp (single f) k in
    let k = ctrl (single f) k in
    let k = seq (cons dpr (cons ctrl [])) f k in
    let k = seq (cons dpr (cons dp [])) f k in
    let k = seq (cons dpw (cons poswr [])) f k in
    let k = seq (cons dpw (cons poswr (cons dp []))) f k in
    let k = seq (cons dpw (cons poswr (cons ctrl []))) f k in
(*
    let k = seq (cons ctrlw (cons poswr [])) f k in
    let k = seq (cons ctrlw (cons poswr (cons dpr []))) f k in
*)
    k 

  let ppoext = false

  let ppo f k =
    do_ppo f
      (if ppoext then
        do_ppo
          (fun r k -> match r with
          | ER r -> f (ERS [r; Rf Ext]) k
          | ERS rs -> f (ERS (rs@[Rf Ext])) k
          | PPO -> assert false)
          k
      else k)

  let () =
    if !Misc.verbose > 0 then begin
      eprintf "PPO is:\n" ;
      ppo (fun r () -> eprintf "%s\n" (R.pp_relax r)) ()
    end 

(*******)


  open C

  let next_reg x = PPC.alloc_reg x

  let next_init st p init loc =
    let rec find_rec = function
      | (PPC.Reg (p0,r0),loc0)::_ when loc0 = loc && p = p0 ->
          r0,init,st
      | _::rem -> find_rec rem
      | [] ->
          let r,st = next_reg st in
          r,(PPC.Reg (p,r),loc)::init,st in
    find_rec init

  let pseudo = List.map (fun i -> PPC.Instruction i)

  let emit_sta st p init x v =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    init,
    [PPC.Instruction (PPC.Pli (rA,v)); PPC.Macro ("STA",[rA;rB])],
    st

  and emit_sta_idx st p init x idx v =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    init,
    [PPC.Instruction (PPC.Pli (rA,v)); PPC.Macro ("STAX",[rA;idx;rB])],
    st


  let emit_load st p init x =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    rA,init,pseudo [PPC.Plwz (rA,0,rB)],st

  let emit_load_not_zero st p init x =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    let lab = Label.next_label "L" in
    rA,init,
    PPC.Label (lab,PPC.Nop)::
    pseudo
      [PPC.Plwz (rA,0,rB) ; PPC.Pcmpwi (0,rA,0) ; PPC.Pbcc (PPC.Eq,lab)],
    st

  let emit_load_not_eq st p init x rP =
    let rA,st = next_reg st in
    let rC,st = next_reg st in
    let rB,init,st = next_init st p init x in
    let lab = Label.next_label "L" in
    let out = Label.next_label "L" in
    rA,init,
    PPC.Instruction (PPC.Pli (rC,200))::
    (* 200 X about 5 ins looks for a typical memory delay *)
    PPC.Label (lab,PPC.Nop)::
    pseudo
      [
       PPC.Plwz (rA,0,rB) ; PPC.Pcmpw (0,rA,rP) ;
       PPC.Pbcc (PPC.Ne,out) ; PPC.Paddi (rC,rC,-1) ;
       PPC.Pcmpwi (0,rC,0) ; PPC.Pbcc (PPC.Ne,lab) ;
      ]@
    [PPC.Label (out,PPC.Nop)],
    st
    
  let emit_fno  st p init x =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    rA,init, [PPC.Macro ("FNO",[rA;rB])],st

  and emit_fno2  st p init x =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    rA,init, [PPC.Macro ("FNO2",[rA;rB])],st

  and emit_open_fno st p init x =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    let lab = Label.next_label "DIY" in
    rA,init,
    [ PPC.Label (lab,PPC.Instruction (PPC.Plwarx (rA,PPC.r0,rB))) ],
    lab,st

  and emit_close_fno st p init lab r x =
    let rB,init,st = next_init st p init x in
    init,
    pseudo [PPC.Pstwcx (r,PPC.r0,rB); PPC.Pbcc (PPC.Ne,lab)],
    st

  and emit_store st p init x v =
    if do_sta then emit_sta  st p init x v
    else
      let rA,st = next_reg st in
      let rB,init,st = next_init st p init x in
      init,pseudo [PPC.Pli (rA,v) ; PPC.Pstw (rA,0,rB)],st

  and emit_load_idx st p init x idx =
    let rA,st = next_reg st in
    let rB,init,st = next_init st p init x in
    rA,init,pseudo [PPC.Plwzx (rA,idx,rB)],st

  and emit_store_idx st p init x idx v =
    if do_sta then emit_sta_idx  st p init x idx v
    else
      let rA,st = next_reg st in
      let rB,init,st = next_init st p init x in
      init,pseudo [PPC.Pli (rA,v) ; PPC.Pstwx (rA,idx,rB)],st

  let emit_access  st p init e = match e.dir with
  | R ->
      let r,init,cs,st = emit_load st p init e.loc in
      Some r,init,cs,st
  | W ->
      let init,cs,st = emit_store st p init e.loc e.v in
      None,init,cs,st

  let emit_access_dep st p init e r1 =
    let r2,st = next_reg st in
    let c = PPC.Pxor(PPC.DontSetCR0,r2,r1,r1) in
    match e.dir with
    | R ->
        let r,init,cs,st = emit_load_idx st p init e.loc r2 in
        Some r,init, PPC.Instruction c::cs,st
    | W ->
        let init,cs,st = emit_store_idx st p init e.loc r2 e.v in
        None,init,PPC.Instruction c::cs,st

  let emit_access_ctrl st p init e r1 =
    let lab = Label.next_label "LC" in
    let c =
      [PPC.Instruction (PPC.Pcmpw (0,r1,r1));
       PPC.Instruction (PPC.Pbcc (PPC.Eq,lab));
       PPC.Label (lab,PPC.Nop);] in
    match e.dir with
    | R ->
        let r,init,cs,st = emit_load st p init e.loc in
        Some r,init, c@[PPC.Instruction PPC.Pisync]@cs,st
    | W ->
        let init,cs,st = emit_store st p init e.loc e.v in
        None,init,c@cs,st


  let emit_fence f =
    PPC.Instruction
      (match f with
      | PPC.Sync -> PPC.Psync
      | PPC.LwSync -> PPC.Plwsync)

  let stronger_fence = PPC.Sync
end
