(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

module type S = sig
  type fence
  type edge


  type relax =
(* Single edge *)
    | ER of edge
(* Sequence of edges (eg Cumulativity) *)
    | ERS of edge list 
    | PPO

  val ac_fence : fence -> sd -> extr -> extr -> relax
  val bc_fence : fence -> sd -> extr -> extr -> relax
  val bc_dp : sd -> extr -> relax

(* Call function over all reckognized relaxations *)
  val fold_relax : (relax -> 'a -> 'a) -> 'a -> 'a

  val compare : relax -> relax -> int
  val pp_relax : relax -> string
  val pp_relax_list : relax list -> string
  val edges_of : relax -> edge list

(* Replace Irr directions in par expansion to W and R *)
  val expand_relaxs :
      ((relax -> relax list -> relax list) -> relax list -> relax list) ->
        relax list -> relax list


(* parsing *)
  val  parse_relax : LexUtil.t -> relax

(* Sets *)
  module Set : MySet.S with type elt = relax
  val pp_set : out_channel -> Set.t -> unit

  module SetSet : MySet.S with type elt = Set.t
  val pp_set_set : out_channel -> SetSet.t -> unit

end

module Make (F:Fence.S) (E:Edge.S with type fence = F.fence) : S
with type fence = E.fence
and type edge = E.edge
= struct 
  type fence = E.fence
  type edge = E.edge
  type relax =
    | ER of edge
    | ERS of edge list
    | PPO

  let compare r1 r2 = match r1,r2 with
  | ER e1,ERS [e2]
  | ERS [e1],ER e2 ->  Pervasives.compare e1 e2
  | _,_ -> Pervasives.compare r1 r2

  open E

(* Cumulativity macros *)


  let ac_fence f sl d1 d2 = ERS [Rf Ext; Fenced (f,sl,d1,d2)]
  let bc_fence f sl d1 d2 = ERS [Fenced (f,sl,d1,d2); Rf Ext]
  let bc_dp sl d = ERS [Dp (sl,d); Rf Ext]

(* Pretty print, macros are filtered and printed specially *)
  let pp_relax r =  match r with
  | ER e | ERS [e] -> E.pp_edge e
  | ERS [Rf Ext; Fenced _ as e] ->
      sprintf "AC%s" (pp_edge e)
  | ERS [Fenced _ as e; Rf Ext] ->
      sprintf "BC%s" (pp_edge e)
  | ERS [Dp _ as e;Rf Ext] ->
      sprintf "BC%s" (pp_edge e) 
  | ERS es ->
      sprintf "[%s]" (String.concat "," (List.map pp_edge es))
  | PPO -> "PPO"


  let pp_relax_list lr = String.concat " " (List.map pp_relax lr)

(* Fold over all relaxations *)

  let fold_relax f k =
    let k = E.fold_edges (fun e -> f (ER e)) k in
    let k = 
      F.fold_fences
        (fun fe ->
          Code.fold_sd_extr
            (fun sd e k ->
              let k = f (ac_fence fe sd Irr e) k in
              let k = f (ac_fence fe sd (Dir R) e) k in
              let k = f (bc_fence fe sd e Irr) k in
              let k = f (bc_fence fe sd e (Dir W)) k in
              k)) k in
    let k =
      Code.fold_sd
        (fun sd k ->
          let k = f (bc_dp sd Irr) k in
          let k = f (bc_dp sd (Dir W)) k in
          k) k in
    let k = f PPO k in
    k

  let iter_relax f = fold_relax (fun r () -> f r) ()


(***********)
(* Parsing *)
(***********)

(*
  Same idea as for edges: pretty print all relaxations
  so as to build a table of reckognized relaxations.
 *)

(* Lexeme table *)
let t = Hashtbl.create 101

(* Fill up lexeme table *)
let () =
  iter_relax
    (fun e ->
      let pp = pp_relax e in
      Hashtbl.add t pp e);
  ()

  let do_parse_relax s = 
    try ER (parse_edge s) (* Because some edges have special parsing *)
    with _ ->
      try Hashtbl.find t s
      with Not_found ->
        Warn.fatal "Bad relax: %s" s

let edges_of r = match r with
  | ER e  -> [e]
  | ERS es -> es
  | PPO -> assert false

(*************************************************************)
(* Expansion of irrelevant direction specifications in edges *)
(*************************************************************)

  let expand_dir d f = match d with
  | Dir _ -> f d
  | Irr -> fun k -> f (Dir W) (f (Dir R) k)
        
  let expand_dir2 e1 e2 f = 
    expand_dir e1
      (fun d1 -> expand_dir e2 (fun d2 -> f d1 d2))

  let do_expand_edge e f = match e with
  | Rf _ | RfStar _ | Fr _ | Ws _ | Hat -> f e
  | Po(sd,e1,e2) ->
      expand_dir2 e1 e2 (fun d1 d2 -> f (Po(sd,d1,d2)))
  | Fenced(fe,sd,e1,e2) ->
      expand_dir2 e1 e2 (fun d1 d2 -> f (Fenced(fe,sd,d1,d2)))
  | Dp(sd,e) ->  expand_dir e (fun d -> f (Dp(sd,d)))
  | Ctrl(sd,e) ->  expand_dir e (fun d -> f (Ctrl(sd,d)))

  let rec do_expand_edges es f suf = match es with
  | [] -> f suf
  | e::es ->
      do_expand_edge e
        (fun e k ->
          try
            let suf = match suf with
            | [] -> [e]
            | f::_ ->
                if can_precede_dirs e f then e::suf
                else raise Exit in
            do_expand_edges es f suf k
          with Exit -> k)

  let expand_edges es f = do_expand_edges (List.rev es) f []


(* Ouf, maintenat c'est facile *)
        

  let expand_relaxs ppo rs =

    let rec do_expand_relax r f = match r with
    | ER e -> do_expand_edge e (fun e -> f (ER e))
    | ERS es -> expand_edges es (fun es -> f (ERS es))
    | PPO  -> ppo (fun r -> do_expand_relax r f) in
    
    let expand_relax r =  do_expand_relax r Misc.cons in

    List.fold_right expand_relax rs []

  open LexUtil

  let parse_relax = function
    | One r -> do_parse_relax r
    | Seq [] -> Warn.fatal "Empty relaxation"
    | Seq es -> ERS (List.map E.parse_edge es)


(********)
(* Sets *)
(********)

  module Set =
    MySet.Make
      (struct
        type t = relax
        let compare = compare
      end)

  let pp_set chan t =
    fprintf chan "{" ;
    Set.pp chan ", "
      (fun chan r -> fprintf chan "%s" (pp_relax r))
      t ;
    fprintf chan "}"    


  module SetSet = MySet.Make(Set)

  let pp_set_set chan ts = SetSet.pp chan " " pp_set ts

end
