(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

module Make
    (T:Test.S)
    (C:XXXCompile.S with module A = T.A) =
  struct
    open Printf
    module A = T.A
    module V = A.V
    module Constr = T.C
    open A.Out

    let rec extract_pseudo ins = match ins with
    | A.Nop -> StringSet.empty
    | A.Label (_,ins) -> extract_pseudo ins
    | A.Instruction ins -> C.extract_addrs ins
    | A.Macro (_,_) -> assert false


    let extract_addrs code =
      List.fold_right
        (fun ins env ->
          StringSet.union (extract_pseudo ins) env)
        code
        StringSet.empty

    let rec compile_pseudo env ins k = match ins with
    | A.Nop -> k
    | A.Label (lbl,ins) ->
        { memo=sprintf "%s:" (A.Out.dump_label lbl) ; inputs=[] ; outputs=[] ;
          label = Some lbl ; branch=[Next] ; }::
        compile_pseudo env ins k
    | A.Instruction ins -> C.compile_ins ins k
    | A.Macro (_,_) -> assert false         
                                                
    let compile_code env code =
      List.fold_right (compile_pseudo env) code []

        
    module RegSet =
      MySet.Make
        (struct
          type t = A.reg
          let compare = A.reg_compare
        end)

    let pp_reg_set chan rs =
      RegSet.pp chan ","
        (fun chan r -> fprintf chan "%s" (A.pp_reg r))
        rs

    module LabEnv =
      Map.Make
        (struct
          type t = string
          let compare = String.compare
        end)

(* Compute the set of registers that are to be inititialized
   by real code. This is standard live-in calculation *)

(* One instruction *)
    let live_in_ins ins (env,live_in_next) =
      let live_out =
        List.fold_left
          (fun k flow ->
            let rs =  match flow with
            | Next -> live_in_next
            | Branch lbl ->
                try LabEnv.find lbl env
                with Not_found -> RegSet.empty in
            RegSet.union rs k)
          RegSet.empty ins.branch in
(* Wrong if there are store conditional... *)
      let live_in =
        RegSet.union
          (RegSet.of_list ins.inputs)
          (RegSet.diff live_out (RegSet.of_list ins.outputs)) in
      (match ins.label with
      | None -> env
      | Some lbl -> LabEnv.add lbl live_in env),
      live_in

(* One sequence of instruction *)
    let live_in_code code env live_in_final =
      List.fold_right live_in_ins code (env,live_in_final)

(* Fixpoint *)
    let comp_fix code live_in_final =
      let rec do_rec env0 =
        let env,r = live_in_code code env0 live_in_final in
(*        eprintf "FIX: {%a}\n" pp_reg_set r ; *)
        let c =
          LabEnv.compare RegSet.compare env env0 in
        if c = 0 then r
        else do_rec env in
      do_rec LabEnv.empty

    let comp_initset code inputs_final =
      RegSet.elements (comp_fix code inputs_final)

    let compile_init proc init code final =
      let locs = Constr.locations final in
      let inputs_final =
        List.fold_right
          (fun loc k -> match loc with
          | A.Location_reg (p,reg) when p = proc -> RegSet.add reg k
          | _ -> k)
          locs RegSet.empty in
      let inputs = comp_initset code inputs_final in
      List.map
        (fun reg ->
          reg,A.find_in_state (A.Location_reg (proc,reg)) init)
        inputs

    let load_addrs env =
      let lst =
        MapString.fold
          (fun addr i k -> (i,addr)::k)
          env [] in
      List.sort
        (fun (x,_) (y,_) -> Misc.int_compare x y)
        lst

    module LocSet =
      MySet.Make
        (struct
          type t = A.location
          let compare = A.location_compare
        end)

    let uniq xs ys = LocSet.elements (LocSet.of_list (xs@ys))

    let compile_final proc final flocs =
      let locs = uniq (Constr.locations final) flocs in
      List.fold_right
        (fun loc k -> match loc with
        | A.Location_reg (p,reg) when p=proc -> reg::k
        | _ -> k)
        locs []

    let mk_templates init code final flocs =
      let outs =
        List.map
          (fun (proc,code) ->
            let addrs = extract_addrs code in
            let _,env =
              StringSet.fold
                (fun a (n,env) -> n+1,MapString.add a n env)
                addrs (0,MapString.empty) in
            let code = compile_code env code in
            proc,env,code)
          code in
      let flocs = List.map fst flocs in
      List.map
        (fun (proc,env,code) ->
          proc,
          { init = compile_init proc init code final ;
            addrs = load_addrs env;
            final = compile_final proc final flocs;
            code = code; })
        outs

    let typeof v = match v with
    | V.Concrete _ -> T.Int
    | V.Symbolic _ -> T.Pointer

    let add_addr_type a ty env =
      try
        let tz = MapString.find a env in
        let ty =
          match ty,tz with
          | T.Int,T.Int -> T.Int
          | (T.Pointer,_)|(_,T.Pointer) -> T.Pointer in
        MapString.add a ty env
      with
        Not_found -> MapString.add a ty env
            
    let add_value v env = match v with
    | V.Concrete _ -> env
    | V.Symbolic a -> add_addr_type a T.Int env

    let comp_globals init code =
      let env =
        List.fold_right
          (fun (loc,v) env ->
            let env = add_value v env in
            match loc with
            | A.Location_global (a) -> 
                add_addr_type a (typeof v) env
            | _ -> env)
          init MapString.empty in
      let env =
        List.fold_right
          (fun (_,t) ->
            List.fold_right
              (fun (_,a) -> add_addr_type a T.Int)
              t.addrs)
          code env in
      MapString.fold
        (fun a ty k -> (a,ty)::k)
        env []

    let type_in_final p reg final flocs =
      ConstrGen.fold_constr
        (fun (loc,v) t -> match loc with
        | A.Location_reg (q,r) when p=q && A.reg_compare reg r = 0 ->
            begin match typeof v,t with
            | (_,T.Pointer)
            | (T.Pointer,_) -> T.Pointer
            | T.Int,T.Int -> T.Int
            end
        | _ -> t)
        final
        (List.fold_right
           (fun (loc,t) k -> match loc with
           | A.Location_reg (q,r) when p=q && A.reg_compare reg r = 0 ->
               begin match t with
               | MiscParser.I -> k
               | MiscParser.P -> T.Pointer
               end
           | _ -> k) 
           flocs
           T.Int)

    let type_out p t final flocs =
      List.map
        (fun reg -> reg,type_in_final p reg final flocs)
        t.final
        
    let type_outs code final flocs =    
      List.map
        (fun (p,t) -> p,(t,type_out p t final flocs))
        code

    let compile t =
      let
          { MiscParser.init = init ;
            info = info;
            prog = code;
            condition = final; 
	    kinds = quantifiers;
            locations = locs ;
	  } = t in
      let final = T.find_our_constraint final quantifiers in
      let code = mk_templates init code final locs in
      let code_typed = type_outs code final locs in
      { T.init = init;
        info = info;
        code = code_typed;
        condition = final;
        globals = comp_globals init code;
        flocs = List.map fst locs ;
      }
        
  end
