(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

open Printf
open Globals

module Make(T:Test.S) : sig
  val dump : Name.t -> T.t -> out_channel -> unit
end = struct
  module A = T.A
  module C = T.C
  open A.V

(* Location utlities *)        
  open ConstrGen
  module LocSet =
    MySet.Make
      (struct
        type t = A.location
        let compare = A.location_compare
      end)

  let do_profiling = Globals.get_profiling ()
  let do_tsc = Globals.get_tsc ()
  let do_clocktime = Globals.get_clocktime ()
  let do_timing = do_tsc || do_clocktime
  let do_safer = Globals.get_safer A.arch
  let do_preload = Globals.get_preload A.arch
  let do_sync_macro = match Globals.get_sync_macro () with
  | None -> false
  | Some _ -> true

  let sync_macro_n  = match Globals.get_sync_macro () with
  | None -> 0
  | Some n -> n

  let memory = get_memory A.arch
  let do_cache = get_cache ()
  let do_call = get_call ()
  let do_prealloc = get_prealloc ()
  let launch = get_launch A.arch
  let affinity = get_affinity ()

  let do_affinity = match affinity with
  | AffinityNone -> false
  | AffinityIncr _ -> true

  let get_final_locs t =
    LocSet.elements
      (LocSet.of_list (C.locations t.T.condition@t.T.flocs)) 

  let get_finals_globals t =
    List.filter
      (fun loc -> match loc with
      | A.Location_global _ -> true
      | _ -> false)
      (get_final_locs t)

  let find_global_init a t =
    A.find_in_state (A.Location_global a) t.T.init

  let have_finals_globals t = match get_finals_globals t with
  | [] -> false
  | _::_ -> true

  let dump_loc_name loc =  match loc with
  | A.Location_reg (proc,reg) -> A.Out.dump_out_reg  proc reg
  | A.Location_global s -> s
        
  let dump_loc_copy loc = "_" ^ dump_loc_name loc ^ "_i" 

  let dump_loc loc = match loc with
  | A.Location_reg (proc,reg) ->
      sprintf "%s[_i]" (A.Out.dump_out_reg  proc reg)
  | A.Location_global s ->
      begin match memory with
      | Direct ->
          sprintf "%s[_i]" s
      | Indirect ->
          sprintf "*(%s[_i])" s
      end

  let dump_ctx_loc pref loc = match loc with
  | A.Location_reg (proc,reg) ->
      sprintf "%s%s[_i]" pref (A.Out.dump_out_reg  proc reg)
  | A.Location_global s ->
      begin match memory with
      | Direct ->
          sprintf "%s%s[_i]" pref s
      | Indirect ->
          sprintf "*(%s%s[_i])" pref s
      end

(* Dump left value *)
  let dump_leftval a = match memory with
  | Direct -> sprintf "%s[_i]" a
  | Indirect -> sprintf "*(%s[_i])" a

(* Dump left & right values when pointer to context is available *)

(* Left value *)
  let dump_a_leftval a = match memory with
  | Direct -> sprintf "_a->%s[_i]" a
  | Indirect -> sprintf "*(_a->%s[_i])" a

    
(* Right value *)
  let dump_a_addr = match memory with
  | Direct -> sprintf "&(_a->%s[_i])"
  | Indirect -> sprintf "_a->%s[_i]"

  let dump_a_v = function
  | A.V.Concrete i ->  sprintf "%i" i
  | A.V.Symbolic s -> dump_a_addr s

(* Right value, casted if pointer *)
  let dump_a_v_casted = function
  | A.V.Concrete i ->  sprintf "%i" i
  | A.V.Symbolic s -> sprintf "((int *)%s)" (dump_a_addr s)

(* Dump left & right values when context is available *)

(* Left value *)
  let dump_ctx_leftval a = match memory with
  | Direct -> sprintf "ctx.%s[_i]"
  | Indirect -> sprintf "ctx.mem_%s[_i]"

    
(* Right value from ctx *)
  let dump_ctx_addr = match memory with
  | Direct -> sprintf "&(ctx.%s[_i])"
  | Indirect -> sprintf "ctx.%s[_i]"

  let dump_ctx_v = function
  | A.V.Concrete i ->  sprintf "%i" i
  | A.V.Symbolic s -> dump_ctx_addr s

(* Idem casted *)

  let dump_ctx_v_casted = function
  | A.V.Concrete i ->  sprintf "%i" i
  | A.V.Symbolic s -> sprintf "((int *)%s)" (dump_ctx_addr s)

  let build_env test =
    List.map
      (fun (s,t) -> A.Location_global s,t)
      test.T.globals @
    List.flatten
      (List.map
         (fun (proc,(_,outs)) ->
           List.map
             (fun (reg,t) -> A.Location_reg (proc,reg),t)
             outs)
         test.T.code)

  let rec find_type loc env = match env with
  | [] -> assert false
  | (loc0,t)::env ->
      if A.location_compare loc loc0 = 0 then t
      else find_type loc env

  let find_global_type a = find_type (A.Location_global a)

  let indent = "  "
  let indent2 = indent ^ indent
  let indent3 = indent2 ^ indent
  let indent4 = indent3 ^ indent
  let indent5 = indent4 ^ indent
  let indent6 = indent5 ^ indent

  let dump_header chan test =
    MySys.output_line chan "/* Parameters */" ;
    Globals.dump_bench A.arch chan ;
    let n = (List.length test.T.code) in
    fprintf chan "#define N %i\n" n ;
     begin match affinity with
     | AffinityIncr i ->
         fprintf chan "#define AFF_INCR %i\n" i
     |  AffinityNone ->
         fprintf chan "#define AFF_INCR (-1)\n"
     end ;
    output_string chan
      "/* Includes */
#include <stdio.h>\n\
#include <stdlib.h>\n\
#include <unistd.h>\n\
#include <errno.h>\n\
#include <assert.h>\n\
#include <time.h>\n\
#include <limits.h>\n\
#include \"utils.h\"\n\
#include \"outs.h\"\n" ;
    if do_affinity then output_string chan "#include \"affinity.h\"\n" ;
    output_string chan
"
/* params */
static int verbose ;
static int size_of_test = SIZE_OF_TEST ;
" ;
    if get_speedcheck () then begin
      MySys.output_line chan "static int speedcheck = 1 ;" ;
      MySys.output_line chan "static int stop_now = 0 ;" ;
      ()
    end ;
    fprintf chan "#define SYNC_K %i\n" Globals.sync_const ;
    fprintf chan "#define SYNC_N %i\n" sync_macro_n ;
    if  do_sync_macro then begin
      fprintf chan "static int max_idx = (SYNC_N * SYNC_K);\n"
    end ;
    ()

  let iter_procs chan f test =
    output_string chan
      (String.concat ", "
         (List.map
            (fun (proc,_) -> f proc)
            test.T.code))


(******************)
(* Read time base *)
(******************)

  let dump_rdtsc chan =
    if do_tsc || do_profiling then begin
      let fx indent fmt =
        ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
      let f fmt = fx "" fmt in
      let fi fmt = fx indent fmt in
      f "/******************/" ;
      f "/* Read time base */" ;
      f "/******************/" ;
      f "" ;
      f "inline static tsc_t rdtsc(void) {" ;
      begin match A.arch with
      | Misc.X86 ->
          fi "unsigned int a,d;" ;
          fi "asm volatile(\"rdtsc\" : \"=a\" (a), \"=d\" (d));" ;
          fi "return ((tsc_t)a) | (((tsc_t)d)<<32);"
      | Misc.PPC ->
          fi "tsc_t r;" ;
          fi "%s" "asm volatile(\"mftb %[r1]\" :[r1] \"=&r\" (r));" ;
          fi "return r;"
      | Misc.ARM -> Warn.fatal "ARM time base not implemented"
      end ;
      f "}" ;
      f "" ;
      if do_profiling then begin
        f "/* Resort to get_clocktime */" ;
        f "inline static tsc_t prof1(void) {" ;
        fi "struct timespec ts;" ;
        fi "if (clock_gettime (CLOCK_MONOTONIC, &ts)) err(1,\"clock_gettime\");" ;
        fi "return 1000000000LL * ts.tv_sec + ts.tv_nsec;" ;
        f "}" ;
        f "" ;        
        f "/* Processor time, per thread, only safe way to to? */" ;
        f "inline static tsc_t prof2(void) {" ;
        fi "struct timespec ts;" ;
        fi "if (clock_gettime (CLOCK_THREAD_CPUTIME_ID, &ts)) err(1,\"clock_gettime\");" ;
        fi "return 1000000000LL * ts.tv_sec + ts.tv_nsec;" ;
        f "}" ;
        f ""
      end
    end

(*************************************************)
(* Soft barriers: to synchronize test iterations *)
(*************************************************)

(*
  This barrier is very sensitive to machine load.

  1- With unlimited spinning running time becomes huge when
     there are more threads than processors.
     test on conti (2 procs..)
  2- with settings 1000 * 200 for instance on conti, running
     times become erratic. Maybe because a daemon running literally
     stops the test: one proc is runnable while the other is spinning.
     With higher sizes, running times are less variable from one
     run to the other.

     Running the test under low priority (nice) also demonstrates the effect:
     running time doubles.

 However this is the best barrier we have, in normal conditions,
 especially with high SIZE_OF_TEST parameter. A benefit of
 such a high setting resides in pthread costs (lauch/join/fst_barrier) being
 amortized over a larger number of tests. It also often increases outcome
 variety, maybe by introduction of capacity cache misses.

  Limiting spinning is a BAD idea, it hinders outcome variety, even in
  normal load conditions.
 *)

let user_barrier_def =
"/* Barriers macros */
#define __busy_wait(b) {while ((b) == 0) {} ; }
#define barrier_wait(my_id,i,b) { if ( (i % N) != my_id ) {__busy_wait(b)} else {(b) = 1;}; }

"

let dump_user_barrier_vars chan indent =
  fprintf chan "%sint volatile *barrier;\n" indent

(*******************************************)
(* Pthread utilies, included hard barriers *)
(*******************************************)

  let pthread_barrier_def = DefString.pthread_barrier_def

  let dump_pthread_barrier_vars chan indent =
    fprintf chan "%sbarrier_t *barrier;\n" indent

  let dump_barrier_vars chan ident =
  MySys.output_line chan "/* Barrier for litmus loop */" ;
  match get_barrier A.arch with
  | NoBarrier -> ()
  | Pthread -> dump_pthread_barrier_vars chan indent
  | User -> dump_user_barrier_vars chan indent

  let barrier_def chan =
    begin match get_barrier A.arch with
    | User ->
        output_string chan user_barrier_def ;
        ()
    | Pthread ->
        output_string chan pthread_barrier_def ;
        ()
    | NoBarrier -> ()
   end

(* Threads, very little *)
      
  let dump_threads chan test =
    barrier_def chan ;
    ()



(*************)
(* Variables *)
(*************)

  let dump_type t = match t with
  | T.Int -> "int"
  | T.Pointer -> "int*"

  let dump_global_type t = match memory with
    | Direct -> dump_type t
    | Indirect -> sprintf "%s*" (dump_type t)

  let dump_vars chan indent test =
    let f fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    MySys.output_line chan "/* Shared variables */" ;
    List.iter
      (fun (s,t) ->
        f "%s *%s;" (dump_global_type t) s)
      test.T.globals ;
    begin match memory with
    | Direct -> ()
    | Indirect ->
        List.iter
          (fun (a,t) -> f "%s *mem_%s ;" (dump_type t) a)
          test.T.globals ;
       f "int *_idx;"
   end ;
   ()

  let is_ptr = function
    | T.Pointer -> true
    | T.Int -> false

  let ptr_in_outs test =
    List.exists
      (fun (_,(_,outs)) ->
        List.exists (fun (_,t) -> is_ptr t) outs)
      test.T.code
  ||
    List.exists (fun (_,t) -> is_ptr t) test.T.globals

  let iter_outs f proc = List.iter (f proc)

  let iter_all_outs f test =
    List.iter
      (fun (proc,(_,outs)) -> iter_outs f proc outs)
      test.T.code

  let dump_out_vars chan indent test =
    MySys.output_line chan "/* Final content of observed  registers */" ;
    iter_all_outs
       (fun proc (reg,t) ->
         fprintf chan "%s%s *%s;\n"
           indent (dump_type t)
           (A.Out.dump_out_reg proc reg))
        test

   let fmt_outcome locs env =
    let pp_loc loc =  match loc with
    | A.Location_reg (proc,reg) ->
        sprintf "%i:%s" proc (A.pp_reg reg)
    | A.Location_global s -> sprintf "%s" s in

    let pp_fmt loc = match find_type loc env with
    | T.Int -> "%i"
    | T.Pointer -> "%s" in

    String.concat " "
      (List.map
         (fun loc -> sprintf "%s=%s;" (pp_loc loc) (pp_fmt loc))
         locs)

  let dump_defs_outs chan doc env test =
    (* If some of the output registers is of pointer type,
       we need a special function to print addresses *)
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let f fmt = fx "" fmt in
    let fi fmt = fx indent fmt in

    if ptr_in_outs test then begin
    if false then begin
      (* Pretty print of addresses, obsolete if output print disabled *)
      f "/* Addresses */" ;
      let dump_test = match memory with
      | Direct ->
          fun s ->
            fi "else if (v_addr == (void *)&(_a->%s[_i])) return \"%s\";"
              s s
      | Indirect ->
          fun s ->
            fi "else if (v_addr == (void *)_a->%s[_i]) return \"%s\";"
              s s in
      f "static char *str_addr(ctx_t *_a,int _i,void *v_addr) {" ;
      fi "if(v_addr == NULL) return \"NULL\";" ;
      List.iter (fun (s,_) -> dump_test s) test.T.globals ;
      fi "else return \"???\";" ;
      fprintf chan "}\n\n"
    end ;

(*  Translation to indices *)
      let dump_test k = match memory with
      | Direct ->
          fun s ->
            fi "else if (v_addr == (void *)&(_a->%s[_i])) return %i;"
              s k
      | Indirect ->
          fun s ->
            fi "else if (v_addr == (void *)_a->%s[_i]) return %i;"
              s k in
      f "static int idx_addr(ctx_t *_a,int _i,void *v_addr) {" ;
      fi "if(v_addr == NULL) { err(2,\"NULL\"); return -1;}" ;
      Misc.iteri (fun k (s,_) -> dump_test k s) test.T.globals ;
      fi "else { err(2,\"???\"); return -1;}" ;
      f "}\n" ;
(* Pretty-print indices *)
      let naddrs = List.length test.T.globals in
      fprintf chan "\nstatic char *pretty_addr[%i] = {" naddrs ;
      List.iter
        (fun (s,_) -> fprintf chan "\"%s\", " s)
        test.T.globals ;
      fprintf chan "} ;\n\n"
    end ;
(* Outcome collection *)
    output_string chan "/**********************/\n" ;
    output_string chan "/* Outcome collection */\n" ;
    output_string chan "/**********************/\n\n" ;
(* Outcome type definition *)
    let outs = get_final_locs test in
    let nouts = List.length outs in
    f "#define NOUTS %i" nouts ;
    f "typedef int outcome_t[NOUTS];" ;
    f "" ;
    Misc.iteri
      (fun i loc ->
        f "static const int %s_f = %i ;" (dump_loc_name loc) i)
      outs ;
    f "" ;
(* Constant wrappers *)    
    output_string chan DefString.hist_defs ;
(* Dumping *)
    f "static void do_dump_outcome(FILE *fhist, int *o, count_t c) {" ;
    let fmt_var = fmt_outcome outs env in
    let fmt ="\"%-6\"PCTR\":>" ^ fmt_var ^ "\\n\"" in
    let args =
      String.concat ","
        ("c"::
         List.map
           (fun loc ->
             let sloc = dump_loc_name loc in
             match find_type loc env with
           | T.Int -> sprintf "o[%s_f]" sloc
           | T.Pointer -> sprintf "pretty_addr[o[%s_f]]" sloc)
           outs) in
    fi "fprintf(fhist,%s,%s);" fmt args ;
    f "}" ;
    f "" ;
    f "static int dump_hist(FILE *fhist,hist_t *h) {" ;
    let c = test.T.condition in
    fi "fprintf(fhist,\"Test %s %s\\n\") ;"
      doc.Name.name (pp_kind (kind_of c));      
    fi "fprintf(fhist,\"%s\",finals_outs(h->outcomes)) ;"
      "Histogram (%\"PCTR\" states)\\n" ;
    fi "outcome_t buff ;" ;
    fi "dump_outs(fhist,do_dump_outcome,h->outcomes,buff,NOUTS) ;" ;
    let to_check =
      if ConstrGen.is_existential c then "h->n_pos > 0"
      else "h->n_neg == 0" in
    fi "int cond = %s;" to_check ;
    fi "fprintf(fhist,\"%%s\\n\",%s);" "cond?\"Ok\":\"No\"" ;
    fi "fprintf(fhist,\"\\nWitnesses\\n\");" ;
    fi "fprintf(fhist,\"Positive: %%\"PCTR\", Negative: %%\"PCTR\"\\n\",h->n_pos,h->n_neg);" ;

    let pp_atom (loc,v) = sprintf "%s=%s" (A.pp_location loc) (A.V.pp_v v) in
    fi "fprintf(fhist,\"Condition %s is %%svalidated\\n\",%s);"
      (String.escaped
         (ConstrGen.constraints_to_string pp_atom c))
      (sprintf "%s ? \"\" : \"NOT \"" to_check) ;
  fi "return cond;" ;
  f "}" ;
  f "" ;
  ()

(* Timing definitions *)

  let dump_timing_vars chan indent =
    if do_timing then begin
      fprintf chan "/* Timing data */\n" ;
      if do_tsc then
        fprintf chan "%stsc_t *thr_tsc[N];\n" indent ;
      if do_clocktime  then
        fprintf chan "%stsc_t thr_clocktime[N];\n" indent ;
      ()
    end

  let dump_defs_timing chan test = 
    if do_timing then begin
      output_string chan 
"tsc_t get_time_diff (struct timespec later,struct timespec earlier) {
  return 1000000000LL * (later.tv_sec - earlier.tv_sec) + (later.tv_nsec - earlier.tv_nsec);
}\n\n" ;
    end

(* Profiling *)
  let dump_profiling_vars chan ident =
    if do_profiling then begin
      fprintf chan "/* Profiling */\n" ;
      fprintf chan "%stsc_t _A[N],_T[N];\n" indent ;
      fprintf chan "%stsc_t _AC[N],_TC[N];\n" indent ;
      if do_preload || do_safer then begin
        fprintf chan "%stsc_t _C[N];\n" indent ;
        fprintf chan "%stsc_t _CC[N];\n" indent
      end ;
      if do_safer then begin
        fprintf chan "%stsc_t _S[N];\n" indent ;
        fprintf chan "%stsc_t _SC[N];\n" indent
      end
    end

(* Loops *)
  let loop_test_prelude chan indent =
    fprintf chan "%sfor (int _i = size_of_test-1 ; _i >= 0 ; _i--) {\n" indent
      
  and loop_test_postlude chan indent =
    fprintf chan "%s}\n" indent

  let loop_proc_prelude chan indent =
    fprintf chan "%sfor (int _p = N-1 ; _p >= 0 ; _p--) {\n" indent
      
  and loop_proc_postlude chan indent =
    fprintf chan "%s}\n" indent

  and choose_proc_prelude chan indent =
      fprintf chan
        "%sfor (int _p = NT-1 ; _p >= 0 ; _p--) {\n" indent
      
(* Safer more, checking globals *)
  let dump_check_vars chan indent env test =
    if do_safer || do_preload then begin
      let f fmt =
        ksprintf
          (fun s -> fprintf chan "%s%s\n" indent s)
          fmt in
      MySys.output_line chan "/* Check data */" ;
      f "%s" "pb_t *fst_barrier;" ;
      begin match get_finals_globals test with
      | [] -> ()
      | locs ->
          if do_safer then begin
            f "%s" "po_t *s_or ;" ;
            List.iter
              (fun loc ->
                f "%s* cpy_%s[N] ;"
                 (dump_type (find_type loc env)) (dump_loc_name loc))
              locs
          end                                   
    end  
  end

  let dump_check_globals chan env test =
    if do_safer || do_preload then begin
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let f fmt = fx "" fmt in
    let fi fmt = fx indent fmt in
    let fii fmt = fx indent2 fmt in
    let fiii fmt = fx indent3 fmt in



(* CHECKGLOBALS *)
    f "/**************************************/" ;
    f "/* Prefetch (and check) global values */" ;
    f "/**************************************/" ;
    f "" ;
    f "static void check_globals(ctx_t *_a) {" ;

(*LOOP*)
    List.iter
      (fun (a,t) -> match memory with
        | Indirect -> 
           if t=T.Int then fi "int *mem_%s = _a->mem_%s;" a a
        | Direct ->
           fi "%s *%s = _a->%s;" (dump_global_type t) a a)
      test.T.globals ;

    loop_test_prelude chan indent ;

    let dump_test chan (a,t) =
      let v = find_global_init a test in
      match memory,t with
      | Indirect,T.Int->
          fprintf chan "mem_%s[_i] != %s" a  (A.Out.dump_v v)
      | (Indirect,T.Pointer) ->
          fprintf chan "%s != %s"
            (dump_a_leftval a) (dump_a_v_casted v)
      | (Direct,_) ->
          fprintf chan "%s != %s"
            (dump_leftval a) (dump_a_v_casted v) in
    List.iter
      (fun x ->
        fprintf chan "%sif (%s%a) err(2,\"check_globals failed\");\n"
          indent2
          (if do_preload then "rand_bit(&(_a->seed)) && " else "")
          dump_test x)
        test.T.globals ;
      loop_test_postlude chan indent ;
(*END LOOP*)
      fi "pb_wait(_a->fst_barrier);" ;
      f "}\n" ;

(* STABILIZE *)
      let locs = get_finals_globals test in
      begin match locs with
      | _::_ when do_safer ->
        f "" ;
        f "static void stabilize_globals(int _id, ctx_t *_a) {" ;
        List.iter
          (fun loc ->
            let a = dump_loc_name loc
            and t = find_type loc env in
            fi "%s *%s = _a->%s;" (dump_global_type t) a a ;
            fi "%s **cpy_%s = _a->cpy_%s;" (dump_type t) a a)
          locs ;
        f "" ;
        fi "pb_wait(_a->fst_barrier); " ;
        fi "for ( ; ; ) {" ;
        loop_test_prelude chan indent2 ;      
        List.iter
          (fun loc ->
             fiii "cpy_%s[_id][_i] = %s;"
               (dump_loc_name loc) (dump_loc loc))
          locs ;   
        loop_test_postlude chan indent2 ;
        fii "po_reinit(_a->s_or);" ;
        fii "int _found;" ;
        fii "%s" "int _nxt_id = (_id+1) % N;" ;
        fii "_found = 0;" ;
        fii "for (int _i = size_of_test-1 ; _i >= 0 && !_found ; _i--) {" ;
        List.iter
         (fun loc ->
           let a = dump_loc_name loc in
           fiii "if (cpy_%s[_id][_i] != cpy_%s[_nxt_id][_i]) { _found = 1; break; }" a a)
         locs ;
        fii "}" ;
        let fmt = "%i: Stabilizing final state!\\n" in
        fii "if (_found) { fprintf(stderr,\"%s\",_id); }" fmt ;
        fii "if (!po_wait(_a->s_or,_found)) return ;" ;
        fi "}" ;
        f "}\n" ;
        ()
     | _ -> ()
     end
    end
        
                               
  let dump_reinit chan test =
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let f fmt  = fx "" fmt in
    let fi fmt  = fx indent fmt in
    let fii fmt  = fx indent2 fmt in


    f "/*******************************************************/" ;
    f "/* Context allocation, freeing and reinitialization    */" ;
    f "/*******************************************************/" ;
    f "" ;
    begin match memory with
    | Indirect ->
        f "static void shuffle(ctx_t *_a) {" ;
        List.iter
          (fun (a,_) ->
            fi "perm_ints(&(_a->seed),_a->_idx,size_of_test);" ;
            loop_test_prelude chan indent;
            fii "_a->%s[_i] = &(_a->mem_%s[_a->_idx[_i]]);" a a ;
            loop_test_postlude chan indent)
          test.T.globals ;
        f "}" ;
        f ""
    | Direct -> ()
    end ;
(* Initialization, called once *)
    let malloc indent name =
      fi "_a->%s = malloc_check(size_of_test*sizeof(*(_a->%s)));"
        name name in
    f "static void init(ctx_t *_a) {" ;
    begin match memory with
    | Direct -> ()
    | Indirect ->
        List.iter
          (fun (a,_) -> malloc indent (sprintf "mem_%s" a))
          test.T.globals ;
        malloc indent "_idx" ;
        loop_test_prelude chan indent ;
        fii "_a->_idx[_i] = _i;" ;
        loop_test_postlude chan indent
    end ;
    fi "_a->seed = rand();" ;
    List.iter (fun (a,_) -> malloc indent a) test.T.globals ;
    iter_all_outs
       (fun proc (reg,_) -> malloc indent (A.Out.dump_out_reg proc reg))
        test ;
    begin match get_barrier A.arch with
    | Pthread ->
        fi "_a->barrier = barrier_create();"
    | User|NoBarrier -> ()
    end ;

    if do_safer || do_preload then  begin
      fi "_a->fst_barrier = pb_create(N);" ;
      match get_finals_globals test with
      | _::_ as locs when do_safer ->
          fi "_a->s_or = po_create(N);" ;
          loop_proc_prelude chan indent ;
          List.iter
            (fun loc ->
              malloc indent2 (sprintf "cpy_%s[_p]" (dump_loc_name loc)))
          locs ;
          loop_proc_postlude chan indent
      | _ -> ()
    end ;
    begin match get_barrier A.arch with
    | NoBarrier -> ()
    | Pthread -> fi "_a->barrier = barrier_create();"
    | User -> malloc indent "barrier"
    end ;
    if do_tsc then begin
      loop_proc_prelude chan indent ;
      malloc indent2 "thr_tsc[_p]" ;
      loop_proc_postlude chan indent
    end ;
    if do_profiling then begin
      loop_proc_prelude chan indent ;
      fii "_a->_A[_p] = _a->_T[_p] = 0;" ;
      fii "_a->_AC[_p] = _a->_TC[_p] = 0;" ;
      if do_safer || do_preload then begin
         fii "_a->_C[_p] = 0;" ;
         fii "_a->_CC[_p] = 0;"
      end ;
      if do_safer then begin
        fii "_a->_S[_p] = 0;" ;      
        fii "_a->_SC[_p] = 0;"
      end ;
      loop_proc_postlude chan indent ;
      ()
    end ;
    if do_sync_macro then begin
      fi "_a->_scratch = malloc_check(max_idx*sizeof(*(_a->_scratch)));"
    end ;
    f "}\n" ;

(* Finalization, called once *)
    let free indent name = fprintf chan "%sfree((void *)_a->%s);\n" indent name
    and pb_free name = fprintf chan "%spb_free(_a->%s);\n" indent name
    and po_free name = fprintf chan "%spo_free(_a->%s);\n" indent name in

    f "static void finalize(ctx_t *_a) {" ;
    begin match memory with
    | Direct -> ()
    | Indirect ->
        List.iter
          (fun (a,_) -> free indent (sprintf "mem_%s" a))
           test.T.globals ;
        free indent "_idx"
    end ;
    List.iter (fun (a,_) -> free indent a) test.T.globals ;
    iter_all_outs
      (fun proc (reg,_) -> free indent (A.Out.dump_out_reg proc reg))
      test ;
    if do_safer then  begin
      pb_free "fst_barrier" ;
      match get_finals_globals test with
      | [] -> ()
      | locs -> 
          po_free "s_or" ;
          loop_proc_prelude chan indent ;
          List.iter
            (fun loc ->
              free indent2 (sprintf "cpy_%s[_p]" (dump_loc_name loc)))
            locs ;
          loop_proc_postlude chan indent
    end ;
    begin match get_barrier A.arch with
    | NoBarrier -> ()
    | Pthread -> fi "barrier_free(_a->barrier);"
    | User -> free indent "barrier"
    end ;
    if do_tsc then begin
      loop_proc_prelude chan indent ;
      free indent2 "thr_tsc[_p]" ;
      loop_proc_postlude chan indent
    end ;
    if do_sync_macro then free indent "_scratch" ;
    f "}\n" ;

(* Re-initialization, called once per run *)
    output_string chan "static void reinit(ctx_t *_a) {\n" ;
    begin match memory with
    | Indirect -> fi "shuffle(_a);"
    | Direct -> ()
    end ;
    loop_test_prelude chan indent ;
    List.iter
      (fun (a,t) ->
        let v = A.find_in_state (A.Location_global a) test.T.init in
        match t,memory with
        | T.Int,Indirect ->
            fprintf chan "%s_a->mem_%s[_i] = %s;\n"
               indent2 a (dump_a_v v)
        | _,_ ->
           fprintf chan "%s%s = %s;\n"
             indent2 (dump_a_leftval a) (dump_a_v_casted v))
      test.T.globals ;
    begin if do_safer then
      List.iter
        (fun (proc,(_,outs)) ->
          List.iter
            (fun (reg,t) ->
              fprintf chan "%s_a->%s[_i] = %s;\n"
                indent2
                (A.Out.dump_out_reg proc reg)
                (match t with
                | T.Int -> "-239487" (* Susmit's sentinel *)
                | T.Pointer -> "NULL"))
            outs)
        test.T.code ;
    end ;
    begin match get_barrier A.arch with
    | User ->
        fprintf chan "%s_a->barrier[_i] = 0;\n" indent2
    | Pthread|NoBarrier -> ()
    end ;
    fprintf chan "%s}\n" indent ;
    if do_timing then begin
      loop_proc_prelude chan indent ;
      if do_tsc then begin
        loop_test_prelude chan indent2 ;
        fprintf chan "%s_a->thr_tsc[_p][_i] = 0;\n" indent3 ;
        loop_test_postlude chan indent2
      end ;
      if do_clocktime then
        fprintf chan "%s_a->thr_clocktime[_p] = 0;\n" indent2 ;
      loop_proc_postlude chan indent
    end ;
    output_string chan "}\n\n"
      
      
  let dump_mbar chan indent =
    if not (get_safer A.arch) then begin
      let bar = match A.arch with
      | Misc.PPC -> "sync"
      | Misc.X86 -> "mfence"
      | Misc.ARM -> "dmb" in
      fprintf chan "%sasm volatile(\"%s\");\n" indent bar
    end

  let dump_templates chan env test =
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let f fmt = fx "" fmt in
    let fi fmt = fx indent fmt in
    let fii fmt = fx indent2 fmt in
(*
    let fii fmt = fx indent2 fmt in 
    f "#ifdef L2" ;
    f "int volatile _t[L2];" ;
    f "int flood_cache(void) {" ;
    fi "int r=0,k;" ;
    fi "for (k=0 ; k < L2 ; k++) r += _t[k];" ;
    fi "return r;" ;
    f "}\n" ;
    f "#endif" ;
    f "" ;
*)  
    f "/***************/" ;
    f "/* Litmus code */" ;
    f "/***************/" ;
    f "" ;
    f "typedef struct {" ;
    fi "int th_id; /* I am running on this thread */" ;
    if do_affinity then fi "int cpu; /* On this cpu */" ;
    fi "ctx_t *_a;   /* In this context */" ;
    f "} parg_t;" ;
    f "" ;
    List.iter
      (fun (proc,(out,outregs)) ->
        fprintf chan "static void *P%i(void *_vb) {\n" proc ;
        fi "parg_t *_b = (parg_t *)_vb;" ;
        fi "ctx_t *_a = _b->_a;" ;
        if do_profiling then begin
          fi "tsc_t _ta = prof1();" ;
          fi "tsc_t _tac = prof2();"
        end ;
        if do_affinity then fi "write_one_affinity(_b->cpu);" ;
        if do_safer || do_preload then begin
          if do_profiling then begin
            fi "tsc_t _tc = prof1();" ;
            fi "tsc_t _tcc = prof2();"
          end ;
          fi "check_globals(_a);" ;
          if do_profiling then begin
            fi "_a->_C[%i] += prof1()-_tc;" proc ;       
            fi "_a->_CC[%i] += prof2()-_tcc;" proc
          end
        end ;
        begin match get_barrier A.arch with
        | User ->
            fi "int _th_id = _b->th_id;" ;
            fi "int volatile *barrier = _a->barrier;"
        | Pthread ->
            fi "barrier_t *barrier = _a->barrier;"
        | NoBarrier -> ()
        end ;
        List.iter
          (fun a ->
            let t = find_global_type a env in
            fi "%s *%s = _a->%s;" (dump_global_type t) a a)
          (A.Out.get_addrs out) ;
        List.iter
          (fun (r,t) ->
            let name = A.Out.dump_out_reg  proc r in
            fi "%s *%s = _a->%s;" (dump_type t) name name)
          outregs ;
	begin if do_clocktime then fi "struct timespec tp1,tp2;" 
        end; 
	begin if do_tsc then begin
          fi "tsc_t tm1,tm2,*thr_tsc = _a->thr_tsc[%i];" proc;
          end 
        end; 

        begin if do_clocktime then
          fi "clock_gettime (CLOCK_MONOTONIC, &tp1);"
        end;
        if do_profiling then begin
          fi "tsc_t _tt = prof1();" ;
          fi "tsc_t _ttc = prof2();"
        end ;
        loop_test_prelude chan indent ;
        begin match get_barrier A.arch with
        | User ->
            fii "barrier_wait(_th_id,_i,barrier[_i]);"
        | Pthread ->
            fii "barrier_wait(%i,barrier);" proc
        | NoBarrier -> ()
        end ;
	begin if do_tsc then 
          fii "tm1 = rdtsc();"
	end ;
        A.Out.dump chan proc out ;
	begin if do_tsc then 
	  begin
            fii "tm2 = rdtsc(); " ;
	    fii "thr_tsc[_i] = tm2 - tm1;"
	  end
	end ;
        loop_test_postlude chan indent ;
        if do_profiling then begin
          fi "_a->_T[%i] += prof1() - _tt;" proc ;
          fi "_a->_TC[%i] += prof2() - _ttc;" proc
        end ;
        begin if do_clocktime then begin 
          fi "clock_gettime (CLOCK_MONOTONIC, &tp2);" ; 
          fi "_a->thr_clocktime[%i] = get_time_diff (tp2,tp1);\n" proc
          end
        end;
        if do_safer && have_finals_globals test then begin
          if do_profiling then begin
            fi "tsc_t _ts = prof1();" ;
            fi "tsc_t _tsc = prof2();"
          end ;
          fi "stabilize_globals(%i,_a);" proc ;
          if do_profiling then begin
            fi "_a->_S[%i] += prof1() - _ts;" proc ;
            fi "_a->_SC[%i] += prof2() - _tsc;" proc
          end
        end ;
        if do_profiling then begin
          fi "_a->_A[%i] += prof1() - _ta;" proc ;
          fi "_a->_AC[%i] += prof2() - _tac;" proc
        end ;
        fi "return NULL;" ;
        f "}\n")
      test.T.code
      

(* Test condition *)
  let rec dump_prop chan p = match p with
  | Atom(loc,v) ->
      fprintf chan "%s == %s" (dump_loc_copy loc) (dump_ctx_v_casted v)
  | Not p -> fprintf chan "!(%a)" dump_prop p
  | Or [] -> fprintf chan "0"
  | Or [p] -> dump_prop chan p
  | Or (p::ps) ->
      fprintf chan "(%a) || (%a)" dump_prop p dump_prop (Or ps)
  | And [] -> fprintf chan "1"
  | And [p] -> dump_prop chan p
  | And (p::ps) ->
      fprintf chan "(%a) && (%a)" dump_prop p dump_prop (And ps)
  | Implies (p1,p2) ->
      fprintf chan "!(%a) || (%a)" dump_prop p1 dump_prop p2
        
  let dump_condition chan indent c =
    let dump_def p =
      fprintf chan "%scond = %a;\n" indent dump_prop  p in
    match c with
    | ForallStates p
    | QueryState p 
    | ExistsState p ->
        dump_def p
    | NotExistsState p ->
        dump_def p ;
        fprintf chan "%scond = !cond;\n" indent

  let dump_zyva chan doc env test =
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let o s = fx "" "%s" s in
    let oi s = fx indent "%s" s in
    let oii s = fx indent2 "%s" s in
    let f fmt = fx "" fmt in
    let fi fmt = fx indent fmt in
    let fii fmt = fx indent2 fmt in
    let fiii fmt = fx indent3 fmt in
    let fiiii fmt = fx indent4 fmt in
    let procs = List.map (fun (proc,_) -> proc) test.T.code in    

    o "typedef struct {" ;
    oi "int max_run;" ;
    oi "pm_t *p_mutex;" ;
    oi "pb_t *p_barrier;" ;
    if do_prealloc then oi "ctx_t ctx;" ;
    if do_affinity then oi "cpus_t *cpus;" ;
    o "} zyva_t;" ;
    o "" ;
    f "#define NT %s" (if do_call then "(N-1)" else "N") ;
    o "" ;
    o "static void *zyva(void *_va) {" ;
(* Define local vars *)
    oi "zyva_t *_a = (zyva_t *) _va;" ;
    if do_profiling then begin
      oi "tsc_t _TA=prof1(), _TZ=0;" ;
      oi "tsc_t _TAC=prof2(), _TZC=0;"
    end ;
    oi "int max_run = _a->max_run;" ; 
    if do_affinity then oi "write_affinity(_a->cpus);" ;
    oi "pb_wait(_a->p_barrier);" ;
    oi "pthread_t thread[NT];" ;
    oi "parg_t parg[N];" ;
    if do_cache then begin
      oi "op_t *op_arg[NT], *op_ret[NT];" ;
      oi "sarg_t sarg[NT];"
    end ;
    fi "f_t *fun[] = {%s};"
      (String.concat ","
         (List.map (sprintf "&P%i") procs)) ;
    oi "hist_t *hist = alloc_hist();" ;
    begin if do_prealloc then
      oi "ctx_t  ctx = _a->ctx;"
    else
      oi "ctx_t  ctx;"
    end ;
    if do_timing then begin
      let f s =
        fi "tsc_t min_%s[N],max_%s[N],total_%s[N];"
          s s s in
      if do_tsc then f "tsc" ;
      if do_clocktime then f "clocktime" ;
      ()
    end ;
    o "" ;

(* Initialize *)
    if do_timing then begin
      let f s =
        fii "min_%s[_p] = ULLONG_MAX;" s ;
        fii "max_%s[_p] = total_%s[_p] = 0;" s s in
      loop_proc_prelude chan indent ;
      if do_tsc then f "tsc" ;
      if do_clocktime then f "clocktime" ;      
      loop_proc_postlude chan indent ;
      o ""
    end ;
    if not do_prealloc then oi "init(&ctx);" ;

(* Build T preads arguments (which are constant) *)
    loop_proc_prelude chan indent ;
    oii "parg[_p].th_id = _p; parg[_p]._a = &ctx;" ;
    if do_affinity then oii "parg[_p].cpu = _a->cpus->cpu[_p];" ;
    loop_proc_postlude chan indent ;
(* Launch cached threads, if needed *)

    if do_cache then begin
      choose_proc_prelude chan indent ;
      oii "sarg_t *p = &sarg[_p];" ;
      oii "p->max_run = max_run;" ;      
      oii "p->op_arg = op_arg[_p] = op_create();" ;
      oii "p->op_ret = op_ret[_p] = op_create();" ;
      oii "p->arg = &parg[_p];" ;
      oii "launch(&thread[_p],start_thread,p);" ;
      loop_proc_postlude chan indent
    end ;
    o "" ;
(* Loop max_run times *)
   oi "for (int n_run = 0 ; n_run < max_run ; n_run++) {" ;
   oii
     "if (verbose>1) fprintf(stderr, \"Run %i of %i\\r\", n_run, max_run);" ;
   oii "reinit(&ctx);"  ;
(* Start/join threads *)
   if do_profiling then  begin
      oii "tsc_t _t = prof1();" ;
      oii "tsc_t _tc = prof2();"
    end ;

    begin match launch with
    | Changing -> fii "perm_funs(&ctx.seed,fun,N);"
    | Fixed -> ()
    end ;

    choose_proc_prelude chan indent2 ;
    begin if do_cache then begin
      fiii "op_set(op_arg[_p],fun[_p]);" ;
    end else
      fiii "launch(&thread[_p],fun[_p],&parg[_p]);"
    end ;
    loop_proc_postlude chan indent2 ;

    begin match launch with
    | Changing ->
        if do_cache then
          fii "perm_ops(&ctx.seed,op_ret,NT);"
        else
          fii "perm_threads(&ctx.seed,thread,NT);"
    | Fixed -> ()
    end ;

    if do_call then begin
      fii "fun[NT](&parg[NT]);"
    end ;

    choose_proc_prelude chan indent2 ;
    if do_cache then begin
      fiii "op_get(op_ret[_p]);"
    end else begin
      fiii "join(&thread[_p]);"
    end ;
    loop_proc_postlude chan indent2 ;
    if do_profiling then  begin
      oii "_TZ += prof1()-_t;" ;
      oii "_TZC += prof2()-_tc;"
    end ;
(* Log final states *)
    fii "/* Log final states */" ;
    loop_test_prelude chan indent2 ;    
    let locs = get_final_locs test in 

(* Make copies of final locations *)
    List.iter
      (fun loc ->
        fiii "%s %s = %s;"
          (dump_type (find_type loc env))
          (dump_loc_copy loc)
          (dump_ctx_loc "ctx." loc))
      locs ;
    fiii "outcome_t o;" ;
    fiii "int cond;" ;
    f "" ;
(* check globals against stabilized value *)
    if do_safer then begin
      match get_finals_globals test with
      | [] -> ()
      | locs -> begin
         List.iter
           (fun loc ->
             loop_proc_prelude chan indent3 ;
             fiiii
               "if (%s != ctx.cpy_%s[_p][_i]) err(1,\"global %s unstabilized\") ;"
               (dump_loc_copy loc) (dump_loc_name loc) (dump_loc_name loc) ;
             loop_proc_postlude chan indent3)
            locs ;
      end
   end ;
(* Save outcome *)
    List.iter
      (fun loc ->
        fiii "o[%s_f] = %s;"
          (dump_loc_name loc)
          (match find_type loc env with
          | T.Int -> dump_loc_copy loc
          | T.Pointer -> sprintf "idx_addr(&ctx,_i,%s)" (dump_loc_copy loc)))
      locs ;
    fiii "add_outcome(hist,1,o);" ;


(****************)
    dump_condition chan indent3 test.T.condition  ;
    fiii "if (cond) { hist->n_pos++; } else { hist->n_neg++; }" ;

    begin if do_tsc then
      let timefmt = "\"Time taken = " ^ 
		    String.concat ", " (List.map (fun proc -> sprintf "\"PTSC\" cycles(thread %i)" proc) procs) ^ 
		    "\\n\"" in
      let timeargs =
        String.concat ","
          (List.map (sprintf "ctx.thr_tsc[%i][_i]") procs) in
      fiii "if (_a->run) fprintf(_a->run,%s,%s);" timefmt timeargs;
      loop_proc_prelude chan indent3 ;
      fiiii "tsc_t *tscp = ctx.thr_tsc[_p];" ;
      fiiii "total_tsc[_p] += tscp[_i];"  ;
      fiiii "if (tscp[_i] < min_tsc[_p]) min_tsc[_p] = tscp[_i];" ;
      fiiii "if (tscp[_i] > max_tsc[_p]) max_tsc[_p] = tscp[_i];" ;
      loop_proc_postlude chan indent3 ;
      ()
    end;
    loop_test_postlude chan indent2 ;
    begin if do_clocktime then
      let timefmt = "\"Total time taken (for %i runs) = " ^ 
		    String.concat ", " (List.map (fun proc -> sprintf "%%lld ns(thread %i)" proc) procs) ^ 
		    "\\n\"" in
      let timeargs =
        "n_run," ^
        String.concat ","
          (List.map (sprintf "ctx.thr_clocktime[%i]")
             procs) in
      fii "if (run) fprintf(run,%s,%s);" timefmt timeargs ;
      loop_proc_prelude chan indent2 ;
      fiii "tsc_t ct_p = ctx.thr_clocktime[_p];" ;
      fiii "total_clocktime[_p] += ct_p;"  ;
      fiii "if (ct_p < min_clocktime[_p]) min_clocktime[_p] = ct_p;" ;
      fiii "if (ct_p > max_clocktime[_p]) max_clocktime[_p] = ct_p;" ; 
      loop_proc_postlude chan indent2 
    end;
    if get_speedcheck () then begin
      let cond =
        if ConstrGen.is_existential test.T.condition then "hist->n_pos > 0"
        else "hist->n_neg > 0" in
      fii "if (speedcheck && (stop_now || %s)) {" cond ;
      fiii "stop_now = 1;" ;
      if do_cache then begin
        choose_proc_prelude chan indent3 ;
        fiiii "op_set(op_arg[_p],NULL);" ;
        loop_proc_postlude chan indent3
      end ;
      fiii "break;" ;
      fii "}"
    end ;
    fi "}" ;

(* Lots of timing information... *)
    if do_timing then begin
      fi "if (!run) {" ;
      fii "FILE *fhist = stdout;" ;
      f "" ;
      fii "pm_lock(_a->p_mutex); " ;
    end ;

    if do_tsc then begin 
        fii "fprintf(fhist,\"Timing information in cycles:\\n\");" ; 
        List.iter 
          (fun proc -> 
             fprintf chan "%sfprintf (fhist,\"Thread %i cycle count: Max:%%lld; Min:%%lld; Mean:%%lld\\n\",max_tsc[%i],min_tsc[%i],total_tsc[%i]/(size_of_test * max_run));" indent2 proc proc proc proc 
           ) procs ;
        fii "fprintf(fhist,\"\\n\");" 
    end;

    if do_clocktime then begin 
        fii "fprintf(fhist,\"Timing information (clocktime)  for blocks of %%i runs:\\n\",size_of_test);" ; 
        List.iter 
          (fun proc -> 
             fprintf chan "%sfprintf(fhist,\"Thread %i (ns): Max:%%lld; Min:%%lld; Mean:%%lld\\n\",max_clocktime[%i],min_clocktime[%i],total_clocktime[%i]/(max_run));" indent2 proc proc proc proc 
           ) procs;
        fii "fprintf(fhist,\"\\n\");"  ;
        fii "fprintf(fhist,\"And the above, divided by size of block :\\n\");" ; 
        List.iter 
          (fun proc -> 
             fprintf chan "%sfprintf(fhist,\"Thread %i (ns): Max:%%lld; Min:%%lld; Mean:%%lld\\n\",max_clocktime[%i]/size_of_test,min_clocktime[%i]/size_of_test,total_clocktime[%i]/(size_of_test * max_run));" indent2 proc proc proc proc 
           ) procs;
        fii "fprintf(fhist,\"\\n\\n\");" 
    end;
    if do_timing then begin
      fii "pm_unlock(_a->p_mutex); " ;
      oi "}"
    end ;

    o "" ;
    if do_cache then begin
      choose_proc_prelude chan indent ;
      oii "join(&thread[_p]);" ;
      loop_proc_postlude chan indent ;
      choose_proc_prelude chan indent ;
      oii "op_free(op_arg[_p]); op_free(op_ret[_p]);" ;
      loop_proc_postlude chan indent
    end ;
    oi "finalize(&ctx);" ;
    if do_profiling then begin
       oi "_TA = prof1()-_TA;" ;
       oi "_TAC = prof2()-_TAC;" ;
       oi "if (verbose) {" ;
       oii "pm_lock(_a->p_mutex);" ;
       let dump_profiling pref =
         let fmt =
           "Profiling " ^ pref ^ " " ^ doc.Name.name ^
           "\\nTA=\"PTSC\"[%0.2fM], TZ=\"PTSC\"[%0.2fM,%0.2f]\\n" in           
         fii
           "fprintf(stderr,\"%s\",_TA%s,tsc_millions(_TA%s),_TZ%s,tsc_millions(_TZ%s),tsc_ratio(_TZ%s,_TA%s));"
          fmt pref pref pref pref pref pref;
         loop_proc_prelude chan indent2 ;
         let fmt =
           "Thread %i: A=\"PTSC\"[%0.2fM,%0.2f], T=\"PTSC\"[%0.2fM,%0.2f]\\n" in
         let arg =
           String.concat ","
             ("_p"::
             (* ALL *)
             sprintf "ctx._A%s[_p]" pref::
             sprintf "tsc_millions(ctx._A%s[_p])" pref::
             sprintf "tsc_ratio(ctx._A%s[_p],_TZ%s)" pref pref::
              (* Testing time *)
             sprintf "ctx._T%s[_p]" pref::
             sprintf "tsc_millions(ctx._T%s[_p])" pref::
             sprintf "tsc_ratio(ctx._T%s[_p],ctx._A%s[_p])" pref pref::
             []) in
         fiii "fprintf(stderr,\"%s\",%s);" fmt arg ;
         if do_safer || do_preload then begin
           let fmt =
             "          C=\"PTSC\"[%0.2fM,%0.2f]" ^
             (if do_safer then ", S=\"PTSC\"[%0.2fM,%0.2f]\\n"
             else "\\n") in
           let arg =
             String.concat ","
             (* Checking time *)
             (sprintf "ctx._C%s[_p]" pref::
              sprintf "tsc_millions(ctx._C%s[_p])" pref::
              sprintf "tsc_ratio(ctx._C%s[_p],ctx._A%s[_p])" pref pref::
             if do_safer then (* Stabilization time *)
               sprintf "ctx._S%s[_p]" pref::
               sprintf "tsc_millions(ctx._S%s[_p])"pref::
               sprintf "tsc_ratio(ctx._S%s[_p],ctx._A%s[_p])" pref pref::
               []
             else []) in
           fiii "fprintf(stderr,\"%s\",%s);" fmt arg
         end ;
         loop_proc_postlude chan indent2  in

     dump_profiling "" ;
     dump_profiling "C" ;
     oii "pm_unlock(_a->p_mutex);" ;
     oi "}"
    end ;
    oi "return hist;" ;
    o "}\n" ;
    ()


  let dump_def_ctx chan env test =
    let o = MySys.output_line chan in
    o "/**********************/" ;
    o "/* Context definition */" ;
    o "/**********************/" ;
    o "" ;
    o "typedef struct {" ;
    dump_vars chan indent test ;
    dump_out_vars chan indent test ;
    dump_check_vars chan indent env test ;
    dump_barrier_vars chan indent ;
    dump_timing_vars chan indent ;
    dump_profiling_vars chan indent ;
    fprintf chan "%sst_t seed;\n" indent ;
    if do_sync_macro then
      fprintf chan "%schar *_scratch;\n" indent ;
    o "} ctx_t;" ;
  o "" ;
  o "" ;
  ()

  let check_speedcheck chan i f =
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    if get_speedcheck () then begin
       fx i "%s" "if (!speedcheck) {" ;
       f (i^indent) ;
       fx i "}"
    end else f i 

  let dump_main chan doc env test = 
    let fx indent fmt =
      ksprintf (fun s -> fprintf chan "%s%s\n" indent s) fmt in
    let f fmt = fx "" fmt in
    let fi fmt = fx indent fmt in
    let fii fmt = fx indent2 fmt in
    let fiii fmt = fx indent3 fmt in
    let fiiii fmt = fx indent4 fmt in

(* Static list of logical processors *)
    begin match get_logical_procs () with
    | None -> ()
    | Some procs ->
        f "static int logical_procs[] = {%s};" (LexSplit.pp_ints procs) ;
        f "" ;
        ()
    end ;
          
    f "int main(int argc, char **argv) {" ;
(* Define local variables *)
    if do_affinity then begin
      match get_logical_procs () with
      | None ->  fi "cpus_t *def_all_cpus = read_affinity();"
      | Some procs ->
          fi "cpus_t *def_all_cpus = cpus_create_init(%i,logical_procs);"
            (List.length procs)
    end else begin
      fi "cpus_t *def_all_cpus = NULL;"
    end ;
    fi "tsc_t start = timeofday();" ;
    fi "cmd_t def = { 0, NUMBER_OF_RUN, SIZE_OF_TEST, AVAIL, 0, SYNC_N, AFF_INCR, def_all_cpus};" ;
    fi "cmd_t cmd = def;" ;
(* Parse command line *)
    f "" ;
    fi "parse_cmd(argc,argv,&def,&cmd);" ;
    fi "verbose = cmd.verbose; size_of_test = cmd.size_of_test;" ;
    fi "int n_exe;" ;    
    fi "if (cmd.n_exe > 0) {" ;
    fii "n_exe = cmd.n_exe;" ;
    fi "} else {" ;
    fii "n_exe = cmd.avail < N ? 1 : cmd.avail / N;" ;
    fi "}" ;
    fi "if (verbose) {" ;
    if do_affinity then begin
      let fmt = "n=%i, r=%i, s=%i, i=%i, p='" in
      fii "fprintf(stderr, \"%s\", n_exe, cmd.max_run, size_of_test, cmd.aff_incr);" fmt ;
      fii "cpus_dump(stderr,cmd.aff_cpus);" ;
      fii "fprintf(stderr,\"'\\n\");"
    end else begin
      let fmt = "n=%i, r=%i, s=%i\\n" in
      fii "fprintf(stderr, \"%s\", n_exe, cmd.max_run, size_of_test);" fmt
    end ;
    fi "}" ;
    if do_sync_macro then  fi "max_idx = SYNC_K * cmd.sync_n;\n" ;
    if do_affinity then begin
      fi "" ;
      fi "cpus_t *all_cpus = cmd.aff_cpus;"
    end ;

(* Spawn experiments *)
    fi "pthread_t th[n_exe];" ;
    fi "zyva_t zarg[n_exe];" ;
    fi "pm_t *p_mutex = pm_create();" ;
    fi "pb_t *p_barrier = pb_create(n_exe);" ;    

    if do_affinity then begin
      fi "int next_cpu = 0;" ;
      begin match affinity with
      | AffinityNone -> assert false
      | AffinityIncr _ ->
          fi "int delta=cmd.aff_incr;" ;
          fi "if (delta <= 0) {" ;
          fii "for (int k=0 ; k < all_cpus->sz ; k++) all_cpus->cpu[k] = -1;" ;
          fii "delta = 1;" ;
          fi "} else {" ;
          fii "delta %%= all_cpus->sz;" ;
          fi"}" ;
          fi "int start_scan=0, max_start=gcd(delta,all_cpus->sz);" ;
      end
    end ;
    fi "for (int k=0 ; k < n_exe ; k++) {" ;
    fii "zyva_t *p = &zarg[k];" ;
    fii "p->max_run = cmd.max_run; p->p_mutex = p_mutex; p->p_barrier = p_barrier; " ;
    if do_prealloc then fii "init(&p->ctx);" ;
    if do_affinity then begin
      fii "p->cpus = cpus_create(N);" ;
      fii "for (int i=0 ; i < N ; i++) {" ;
      fiii "p->cpus->cpu[i] = all_cpus->cpu[next_cpu];" ;
      fiii "next_cpu += delta; next_cpu %%= all_cpus->sz;" ;
      fiii "if (next_cpu == start_scan) {" ;
      fiiii "start_scan++ ; start_scan %%= max_start;" ;
      fiiii  "next_cpu = start_scan;" ;
      fiii "}" ;
      fii "}"
    end ;
    fii "launch(&th[k],zyva,p);" ;
    fi "}" ;
    if do_affinity then fi "cpus_free(all_cpus);" ;

(* Join *)
    f "" ;
    fi "count_t n_outs = size_of_test; n_outs *= cmd.max_run;" ;
    fi "hist_t *hist = (hist_t *)join(&th[0]);" ;
    if do_affinity then fi "cpus_free(zarg[0].cpus);" ;
    fi "for (int k=1 ; k < n_exe ; k++) {" ;
    fii "hist_t *hk = (hist_t *)join(&th[k]);" ;
    if do_affinity then fii "cpus_free(zarg[k].cpus);" ;
    check_speedcheck chan indent2
    (fun i ->
      f
        "%sif (sum_hist(hk) != n_outs || hk->n_pos + hk->n_neg != n_outs) {" 
        i ;
    fi "%serr(1,\"sum_hist\") ;" i ;
    f "%s}" i) ;
  f
"    hist->n_pos += hk->n_pos ;
    hist->n_neg += hk->n_neg ;
    hist->outcomes = merge_outs(hist->outcomes,hk->outcomes,NOUTS) ;
    free_hist(hk) ;
  }
  pm_free(p_mutex) ;
  pb_free(p_barrier) ;

  n_outs *= n_exe ;" ;
  check_speedcheck chan indent
    (fun i ->
      f
      "%sif (sum_hist(hist) != n_outs || hist->n_pos + hist->n_neg != n_outs) {" 
        i ;
    fi "%serr(1,\"sum_hist\") ;" i;
    f "%s}" i) ;
  fi "int cond = dump_hist(stdout,hist);" ;
  fi "free_hist(hist);" ;

  List.iter
    (fun (k,vs) ->
      if k = "Relax" then
        let fmt = sprintf "Relax %s %%s %%s\\n" doc.Name.name in
        fi "fprintf(stdout,\"%s\",cond ? \"Ok\" : \"No\",\"%s\");"
          fmt (String.escaped vs) ;
      else
      let fmt = "%s=%s\\n" in
      fi "fprintf(stdout,\"%s\",\"%s\",\"%s\");"
        fmt (String.escaped k) (String.escaped vs))
    test.T.info ;
  fi "tsc_t total = timeofday() - start ;" ;
  let fmt = sprintf "Time %s %%.2f\\n"  doc.Name.name in
  fi "fprintf(stdout,\"%s\",total / 1000000.0) ;" fmt ;
  fi "exit(cond ^ cond) ;" ;
  f "}" ;
  ()
      
  let dump doc test chan =
(* Minimal type environemnt *)
    let env = build_env test in
    dump_header chan test ;
    dump_rdtsc chan ;
    dump_threads chan test ;
    dump_def_ctx chan env test ;    
    dump_defs_outs chan doc env test ;
    dump_defs_timing chan test ;
    dump_reinit chan test ;
    dump_check_globals chan env test ;
    dump_templates chan env test ;
    dump_zyva chan doc env test ;
    dump_main chan doc env test
end
      
